#!/usr/bin/perl
# Copyright (c) 2006 Red Hat, Inc. All rights reserved. This copyrighted
# material is made available to anyone wishing to use, modify, copy, or
# redistribute it subject to the terms and conditions of the GNU General
# Public License v.2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# Author: Milan Zazrivec

# Desctiption:
# Basic stop/start smoketest of system services enabled in current runlevel.
# Concept is:
#   1) stop all services from /etc/rc.d/init.d that needs to be stopped
#      in their precedence. Script reports stop failures, kills by signal,
#      scipts that return non-zero value but do not print anything and
#      scripts that do not return to the invocation environment.
#   2) start all services enabled in current runlevel + reports (start failures,
#      signal kills, timeouts)

# status subroutine, idea taken from /etc/rc.d/init.d/functions: status()
# returns 1 for running service, 0 otherwise
sub status
{
    local *PIDOF;
    my $output;

    # first - pidof
    if (open PIDOF, "/sbin/pidof $_[0]|")
    {
        $output = <PIDOF>;
        chomp($output);
        if ($output !~ /^$/)
        {
            return 1;
        }
    }
    # second - /var/run/$basename.pid
    if ( -f "/var/run/$_[0].pid")
    {
        return 1;
    }
    # third - /var/lock/subsys/$basename
    if ( -f "/var/lock/subsys/$_[0]")
    {
        return 1;
    }
    return 0;
}

use warnings;
use strict 'refs';

my $runlevel;
my $rcdir;
my $haltdir = "/etc/rc.d/rc0.d";
my $dir_entry;
my $service;
my $svc;
my $cmd;
my $svc_line;
my $errors = 0; # error count
my $ret; 
my $sig;
my $i;
my $svc_timeout = 60; # in seconds
my $eval_ret;

my @svc_out;
my @start_svc;
my @kill_svc;

local *SVCDIR;
local *SVCOUT;

# list of services we won't stop/start
my %svc_blacklist = ('network' => '1', 'sshd' => '1', 'auditd' => '1',
                     'vpnc' => '1', 'named' => '1', 'vmware' => '1');

# SIGALRM handler -- we'll just die inside eval{} when the service times out
local $SIG{ALRM} = sub { die };

# which runlevel did we boot into? -> $runlevel
if (defined ($_ = `/sbin/runlevel`) and /^.+(\d)$/)
{
    $runlevel = $1;
}
else
{
    die "Could not determine current runlevel\n";
}

# what is the rc?.d directory for $runlevel? -> $rcdir;
$rcdir = "/etc/rc.d/rc".$runlevel.".d";

# we'll read all S?? entries from $rcdir and store then in @start_svc
die "$rcdir: $!\n" unless opendir SVCDIR, $rcdir;
while ($dir_entry = readdir SVCDIR)
{
    next unless -f "$rcdir/$dir_entry";
    if ("$rcdir/$dir_entry" =~ /^$rcdir\/(S\d\d.+)$/)
    {
        push @start_svc, $1;
    }
}
closedir SVCDIR;

# we'll read all K?? entries from $haltdir and store then in @kill_svc
die "$haltdir: $!\n" unless opendir SVCDIR, $haltdir;
while ($dir_entry = readdir (SVCDIR))
{
    next unless -f "$haltdir/$dir_entry";
    if ("$haltdir/$dir_entry" =~ /^$haltdir\/(K\d\d.+)$/)
    {
        push @kill_svc, $1;
    }
}
closedir SVCDIR;

# sort stored services from [SK]01 to [SK]99
@start_svc = sort @start_svc;
@kill_svc  = sort @kill_svc;

# we'll try to stop every running service from /etc/rc.d/init.d
foreach $service (@kill_svc)
{
    next unless $service =~ /^K\d\d(.+)$/;
    $svc = $1;

    # skip blacklisted services
    next if ($svc_blacklist{$svc});

    # we only want to stop services that need to be stopped
    next unless status $svc;

    # we need to set NOLOCALE and LC_MESSAGES environment variables
    # so that initscripts would always print OK/FAILED rather than
    # locale's equivalents. see /etc/init.d/functions

    $cmd = "export LC_MESSAGES=C; export NOLOCALE=1; $haltdir/$service stop|";

    $eval_ret = eval 
    {
        # we may come across initscripts that will never return -> timeout
        # see https://bugzilla.redhat.com/bugzilla/show_bug.cgi?id=208079
        alarm $svc_timeout;
        open SVCOUT, "$cmd" or die "$svc: $!\n";
        @svc_out = <SVCOUT>;
        alarm 0;
        close SVCOUT;
        $ret = $? >> 8; # actual return value
        $sig = $? & 127; # did we receive any signal?
    };
    if (not defined $eval_ret) # if died inside eval, return value not defined
    {
        ++$errors;
        print "[STOP TIMEOUT] Service $svc timeout($svc_timeout seconds)\n";
    }

    # did we catch any signal?
    if ($sig)
    {
        ++$errors;
        print "[STOP SIGNAL] service $svc died from signal ",$SIG{$sig},"\n";
	next;
    }

    # did the service provide any output?
    if (not defined $svc_out[0])
    {
        # we print error iff the service exited with non-zero
        next unless $ret;
        ++$errors;
        print "[STOP UNDEF] service: $svc, exited with: $ret, ".
              "did not provide any output\n";
        next; # skip the rest of the loop
    }

    # check for "red FAILED" and try to catch all output from service
    if ($svc_out[0] =~ /^.+FAILED.+$/ or $ret != 0)
    {
        ++$errors;
        print "[STOP FAIL] service: $svc, exited with: $ret\n";
        for ($i = 1; $i < $#svc_out; ++$i)
        {
            print "> ", $svc_out[$i];
        }
    }
}

foreach $service (@start_svc)
{
    next unless $service =~ /^S\d\d(.+)$/;
    $svc = $1;
 
    # skip blacklisted services
    next if ($svc_blacklist{$svc});

    # same trick with NOLOCALE and LC_MESSAGES
    $cmd = "export LC_MESSAGES=C; export NOLOCALE=1; $rcdir/$service start|";

    $eval_ret = eval {
        # we may come across initscripts that will never return -> timeout
	# see https://bugzilla.redhat.com/bugzilla/show_bug.cgi?id=208079
        alarm $svc_timeout; #set the timeout
        open SVCOUT, "$cmd" or die "$svc: $!\n";
        @svc_out = <SVCOUT>;
        alarm 0;
        close SVCOUT;
	$ret = $? >> 8; # actual return value
	$sig = $? & 127; # did we receive any signal?
    };
    if (not defined $eval_ret) # if died inside eval, return value not defined
    {
        ++$errors;
        print "[START TIMEOUT] Service $svc timeout ($svc_timeout seconds)\n";
    }

    # did we catch any signal?
    if ($sig)
    {
        ++$errors;
        print "[START SIGNAL] service $svc died from signal ",$SIG{$sig},"\n";
        next;
    }

    # did the service provide any output?
    if (not defined($svc_out[0]))
    {
        next unless $ret; # print error iff the service exited with non-zero
	++$errors;
        print "[START UNDEF] service: $svc, exited with: $ret, ".
              "did not provide any output\n";
        next; # skip the rest of the loop
    }

    # check for "red FAILED" and try to catch all output from service
    if ($svc_out[0] =~ /^.+FAILED.+$/)
    {
        ++$errors;
        print "[START FAIL] service: $svc exited with: $ret\n";
        for ($i = 1; $i < $#svc_out; ++$i)
        {
            print "> ", $svc_out[$i];
        }
    }
}

exit $errors;
